from pandas import DataFrame, to_datetime

from weaverbird.backends.pandas_executor.date_extractors import (
    extract_current_day_date,
    extract_first_day_of_iso_week,
    extract_first_day_of_month,
    extract_first_day_of_quarter,
    extract_first_day_of_week,
    extract_first_day_of_year,
)
from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import DateGranularityStep


def execute_date_granularity(
    step: DateGranularityStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    granularity = step.granularity
    new_column_name = step.new_column if step.new_column is not None else step.column

    series_dt = to_datetime(df[step.column], utc=True)

    if granularity == "year":
        result = extract_first_day_of_year(series_dt)
    elif granularity == "month":
        result = extract_first_day_of_month(series_dt)
    elif granularity == "week":
        result = extract_first_day_of_week(series_dt)
    elif granularity == "quarter":
        result = extract_first_day_of_quarter(series_dt)
    elif granularity == "isoWeek":
        result = extract_first_day_of_iso_week(series_dt)
    elif granularity == "day":
        result = extract_current_day_date(series_dt)
    else:
        return df

    return df.assign(**{new_column_name: result})



from weaverbird.backends.mongo_translator.date_extractors import (
    extract_current_day,
    extract_first_day_of_iso_week,
    extract_first_day_of_month,
    extract_first_day_of_quarter,
    extract_first_day_of_week,
    extract_first_day_of_year,
)
from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import DateGranularityStep

_EXTRACT_MAP = {
    "year": extract_first_day_of_year,
    "month": extract_first_day_of_month,
    "week": extract_first_day_of_week,
    "quarter": extract_first_day_of_quarter,
    "isoWeek": extract_first_day_of_iso_week,
    "day": extract_current_day,
}


def translate_date_granularity(step: DateGranularityStep) -> list[MongoStep]:
    column_name = step.new_column if step.new_column is not None else step.column
    return [{"$addFields": {column_name: _EXTRACT_MAP[step.granularity](step.column)}}]



from weaverbird.backends.mongo_translator.steps.types import MongoStep


def truncate_date_to_day(expr: dict | str) -> MongoStep:
    return {
        "$dateTrunc": {
            "unit": "day",
            "date": expr,
        },
    }


def extract_first_day_of_year(column: str) -> MongoStep:
    return {
        "$dateFromParts": {"year": {"$year": f"${column}"}, "month": 1, "day": 1},
    }


def extract_first_day_of_month(column: str) -> MongoStep:
    return {
        "$dateFromParts": {
            "year": {"$year": f"${column}"},
            "month": {"$month": f"${column}"},
            "day": 1,
        },
    }


def extract_first_day_of_week(column: str) -> MongoStep:
    return truncate_date_to_day(
        {
            # We subtract to the target date a number of days corresponding to (dayOfWeek - 1)
            "$subtract": [
                f"${column}",
                {
                    "$multiply": [
                        {"$subtract": [{"$dayOfWeek": f"${column}"}, 1]},
                        24 * 60 * 60 * 1000,
                    ],
                },
            ],
        }
    )


def extract_first_day_of_quarter(column: str) -> MongoStep:
    return {
        "$dateFromParts": {
            "year": {"$year": f"${column}"},
            "month": {
                "$switch": {
                    "branches": [
                        {
                            "case": {"$lte": [{"$divide": [{"$month": f"${column}"}, 3]}, 1]},
                            "then": 1,
                        },
                        {
                            "case": {"$lte": [{"$divide": [{"$month": f"${column}"}, 3]}, 2]},
                            "then": 4,
                        },
                        {
                            "case": {"$lte": [{"$divide": [{"$month": f"${column}"}, 3]}, 3]},
                            "then": 7,
                        },
                    ],
                    "default": 10,
                },
            },
            "day": 1,
        },
    }


def extract_first_day_of_iso_week(column: str) -> MongoStep:
    return truncate_date_to_day(
        {
            # We subtract to the target date a number of days corresponding to (isoDayOfWeek - 1)
            "$subtract": [
                f"${column}",
                {
                    "$multiply": [
                        {"$subtract": [{"$isoDayOfWeek": f"${column}"}, 1]},
                        24 * 60 * 60 * 1000,
                    ],
                },
            ],
        }
    )


def extract_current_day(column: str) -> MongoStep:
    return truncate_date_to_day(f"${column}")
